
from edge import eventlog
from edge.admin.reply_helpers import Reply, REPLY_NOT_FOUND_404, REPLY_SUCCESS_200
from edge.admin.cgi import (run_cgi, CgiModule, CgiPathInfo, CgiRequest,
                            CgiBadRequestError, CgiNotFoundError)


from webserver.alarm.api import *


class _PathInfo(CgiPathInfo):
    def __init__(self, *args, **kwargs):
        CgiPathInfo.__init__(self, *args, **kwargs)

    @property
    def origin(self):
        return self[1]

    @property
    def alarm_name(self):
        return self[2]

    @staticmethod
    def unpack(raw_path_info):
        """
        :type raw_path_info: str
        :rtype: _PathInfo
        """
        # eventlog.log_d('unpack %s' % type(raw_path_info))
        return _PathInfo(raw_path_info.split('/')[1:])


class _Module(CgiModule):
    """
    GET     /api/v2/origins/alarm
    PUT     /api/v2/origins/alarm
    """
    def __init__(self, handler):
        """
        :type handler: () -> Reply
        """
        CgiModule.__init__(self, handler)

    @classmethod
    def unpack(cls, request):
        """
        :type request: CgiRequest
        :rtype: CgiModule

        :raise CgiNotFoundError:
        :raise CgiBadRequestError:
        """
        method_to_handlers_map = {
            'GET': [
                cls._get,
            ],
            'PUT': [
                cls._put
            ],
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request)

            except CgiBadRequestError:
                raise

            except CgiNotFoundError:
                pass

        eventlog.log_i('fail: http404')
        raise CgiNotFoundError

    @staticmethod
    def _get(request):
        """
        :type request: CgiRequest
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 0:
                raise CgiNotFoundError

        def check_args():
            if request.query_string is None:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            origin = None
            name = None
            for name_value in request.query_string[1:].split('&'):
                name, value = name_value.split('=', 1)
                if name == 'name':
                    name = value
                elif name == 'origin':
                    origin = value

            if None in [origin, name]:
                raise CgiBadRequestError

            result = get_alarm(origin, name)
            rc = REPLY_SUCCESS_200 if result else REPLY_NOT_FOUND_404
            return Reply(status=rc)

        return _Module(handler)

    @staticmethod
    def _put(request):
        """
        :type request: CgiRequest
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 2:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            origin = request.path_info[0]
            name = request.path_info[1]
            status = REPLY_SUCCESS_200 if add_alarm(origin, name) else REPLY_NOT_FOUND_404
            return Reply(status=status)

        return _Module(handler)


def _main():
    run_cgi(
        lambda arg: _PathInfo.unpack(arg),
        lambda arg: _Module.unpack(arg))


if __name__ == '__main__':
    eventlog.open_log('cgi-alarm', 'debug')
    try:
        _main()
    finally:
        eventlog.close_log()
