import os

from edge import eventlog
from edge.admin.reply_helpers import Reply
from edge.admin.cgi import (run_cgi, CgiModule, CgiPathInfo, CgiRequest,
                            CgiBadRequestError, CgiNotFoundError,
                            CgiConflictError, CgiInternalServerError)

from webserver.openvpn import api


class _PathInfo(CgiPathInfo):
    def __init__(self, *args, **kwargs):
        CgiPathInfo.__init__(self, *args, **kwargs)

    @property
    def profile(self):
        return self[0]

    @property
    def profile_ref(self):
        return self[1]

    @property
    def active(self):
        return self[2]

    @staticmethod
    def unpack(raw_path_info):
        """
        :type raw_path_info: str
        :rtype: _PathInfo
        """
        # eventlog.log_d('unpack %s' % type(raw_path_info))
        return _PathInfo(raw_path_info.split('/')[1:])


class _Module(CgiModule):
    """
    GET     /api/v2/admin/actions/openvpn
    POST    /api/v2/admin/actions/openvpn  /profiles
    DELETE  /api/v2/admin/actions/openvpn  /profiles/:ref

    POST    /api/v2/admin/actions/openvpn  /profiles/:ref/active
    DELETE  /api/v2/admin/actions/openvpn  /profiles/:ref/active
    """
    def __init__(self, handler):
        """
        :type handler: () -> Reply
        """
        CgiModule.__init__(self, handler)

    @classmethod
    def unpack(cls, request):
        """
        :type request: CgiRequest
        :rtype: CgiModule

        :raise CgiNotFoundError:
        :raise CgiBadRequestError:
        """
        method_to_handlers_map = {
            'GET': [
                cls._get
            ],
            'POST': [
                cls._post_profile,
                cls._post_profile_active
            ],
            'PUT': [

            ],
            'DELETE': [
                cls._delete_profile,
                cls._delete_profile_active
            ],
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request)

            except CgiBadRequestError:
                raise

            except CgiNotFoundError:
                pass

        eventlog.log_i('fail: http404')
        raise CgiNotFoundError

    @staticmethod
    def _get(request):
        """
        :type request: CgiRequest
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 0:
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            reply_jnode = api.get_openvpn_info()
            return Reply(body=reply_jnode)

        return _Module(handler)

    @staticmethod
    def _post_profile(request):
        """
        :type request: CgiRequest
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 1:
                raise CgiNotFoundError

            if path_info.profile != 'profiles':
                raise CgiNotFoundError

        def check_args():
            """
            :rtype: str
            """
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            prof_path = os.environ.get('PROF_PATH')
            if prof_path is None:
                eventlog.log_e('fail: http500 (env.PROF_PATH not found)')
                raise CgiInternalServerError

            prof_name = os.environ.get('PROF_NAME')
            if prof_name is None:
                eventlog.log_e('fail: http500 (env.PROF_NAME not found)')
                raise CgiInternalServerError

            prof_user = os.environ.get('PROF_USER')
            prof_pass = os.environ.get('PROF_PASS')

            return prof_path, prof_name, prof_user, prof_pass

        check_path_info(request.path_info)
        prof_path, prof_name, prof_user, prof_pass = check_args()

        def handler():
            """
            :rtype: Reply
            """
            try:
                reply_jnode = api.add_profile(prof_name, prof_path, prof_user, prof_pass)
                return Reply(body=reply_jnode)
            except api.BadRequestError:
                raise CgiBadRequestError

        return _Module(handler)

    @staticmethod
    def _delete_profile(request):
        """
        :type request: CgiRequest
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 2:
                raise CgiNotFoundError

            if path_info.profile != 'profiles':
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            try:

                reply_jnode = api.delete_profile(request.path_info.profile_ref)
                return Reply(body=reply_jnode)
            except api.NotFoundError:
                raise CgiNotFoundError
            except api.BadRequestError:
                raise CgiBadRequestError
            except Exception:
                eventlog.log_e('fail: delete profile')
                raise CgiInternalServerError

        return _Module(handler)

    @staticmethod
    def _post_profile_active(request):
        """
        :type request: CgiRequest
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 3:
                raise CgiNotFoundError

            if path_info.profile != 'profiles':
                raise CgiNotFoundError

            if path_info.active != 'active':
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body is not None:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            try:
                reply_jnode = api.make_profile_active(request.path_info.profile_ref, True)
                return Reply(body=reply_jnode)
            except api.NotFoundError:
                raise CgiNotFoundError

        return _Module(handler)

    @staticmethod
    def _delete_profile_active(request):
        """
        :type request: CgiRequest
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: _PathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 3:
                raise CgiNotFoundError

            if path_info.profile != 'profiles':
                raise CgiNotFoundError

            if path_info.active != 'active':
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body is not None:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            try:
                path_info = request.path_info  # type: _PathInfo
                reply_jnode = api.make_profile_active(path_info.profile_ref, False)
                return Reply(body=reply_jnode)
            except api.NotFoundError:
                raise CgiNotFoundError

        return _Module(handler)


def _main():
    try:
        run_cgi(
            lambda arg: _PathInfo.unpack(arg),
            lambda arg: _Module.unpack(arg))
    finally:
        pass


if __name__ == '__main__':
    eventlog.open_log('openvpn-cgi', 'debug')
    try:
        _main()
    finally:
        eventlog.close_log()
