#!/usr/bin/python

import base64
import glob
import os
import subprocess
import sys
import ujson
from collections import defaultdict

import edge.admin.ui_types as ui_types
from edge import eventlog
from edge.admin.reply_helpers import *
from edge.pipeline_helpers import silent_mkdir_r

import _helpers
import _vpw

_HASPRUS_FILEPATH = '/opt/edge/bin/hasp_update'
_HASPRUS_KEY_UPDATE = 'u'
_HASPRUS_KEY_INFO = 'i'
_HASPRUS_KEY_FINGERPRINT = 'f'
_HASPRUS_KEY_GETDEMO = 'get_demo'
_HASPRUS_KEY_LIST = 'l'

_DEMO_LICENSE_FILEPATH = '/opt/edge/share/edge-embedded-demo.v2c'
_DEMO_KEY_FILEPATH_TEMPLATE = '/var/hasplm/installed/107392/*_provisional.v2c'


_VENDOR_ID = '107392'
_DEMO_PRODUCT_ID = '1308'
_DEMO_KEY_ID = 'embedded-demo'

_BASE_KEY_NAME_SUFFIX = '_base.v2c'
_UPDATED_KEY_NAME_SUFFIX = '_update*.v2c'
_PROVISIONAL_KEY_NAME_SUFFIX = '_provisional.v2c'

_ATTACHED_KEYS_DIRPATH = '/var/hasplm/installed/%s/' % _VENDOR_ID
_DETACHED_KEYS_DIRPATH = '/etc/edge/hasplm/detached/'

_V2C_TEMP_FILEPATH = '/tmp/license'

_FPS_TO_TITLE_MAP = {
    25: 'license.lpr.group.25',
    6: 'license.lpr.group.6',
    0: 'license.lpr.group.0'
}


def closure(handler, *args, **kwargs):
    def _handler():
        return handler(*args, **kwargs)

    return _handler


def _try_silent_mkdir_r(dirpath):
    """
    :type dirpath: str
    :rtype: bool
    """
    if not os.path.exists(dirpath):
        silent_mkdir_r(dirpath)
        return True

    return False


def _check_key_attached(key_id):
    """
    :type key_id: str
    :rtype: bool
    """
    # mkdir for detached keys when called first time
    _try_silent_mkdir_r(_DETACHED_KEYS_DIRPATH)

    detached_key_filepath_set = glob.glob(_DETACHED_KEYS_DIRPATH + _BASE_KEY_NAME_SUFFIX)

    detached_key_list = map(lambda x: os.path.basename(x.strip(_BASE_KEY_NAME_SUFFIX)), detached_key_filepath_set)

    return key_id not in detached_key_list


def _get_attached_ui_key_jnode_list():
    """
    :rtype: list of dict
    """

    cmdline = ' '.join([_HASPRUS_FILEPATH, _HASPRUS_KEY_LIST])
    hasprus = subprocess.Popen(cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    hasprus_stdout, hasprus_stderr = hasprus.communicate()
    retcode = hasprus.returncode

    if retcode != 0:
        eventlog.log_e('fail: get attached keys list (%s, %s, %s)' % (hasprus_stdout, hasprus_stderr, retcode))
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    local_key_jnode_list = ujson.loads(hasprus_stdout)

    _HASP_HL_CONF = "HASP_HL"

    attached_ui_key_jnode_list = []
    for key_id, configuration in local_key_jnode_list.iteritems():
        local_key_jnode = {
            'id': key_id,
            'removable': not (configuration == _HASP_HL_CONF),
            'detachable': not (configuration == _HASP_HL_CONF),
            'attached': _check_key_attached(key_id)}

        attached_ui_key_jnode_list.append(local_key_jnode)

    demo_key_id = get_demo_key_id()
    if demo_key_id:
        # replace demo-key jnode with stub
        first_demo_ui_key_jnode = filter(
            lambda ui_key_jnode: ui_key_jnode['id'] == demo_key_id,
            attached_ui_key_jnode_list)[0]

        attached_ui_key_jnode_list.remove(first_demo_ui_key_jnode)
        attached_ui_key_jnode_list.append({
            'id': _DEMO_KEY_ID,
            'removable': False,
            'detachable': False,
            'attached': True
        })

    else:
        attached_ui_key_jnode_list.append({
            'id': _DEMO_KEY_ID,
            'removable': False,
            'detachable': True,
            'attached': False
        })

    return attached_ui_key_jnode_list


def _get_detached_ui_key_jnode_list():
    """
    :rtype: list of dict
    """
    detached_ui_key_jnode_list = map(lambda key_id: {
        'id': key_id,
        'removable': True,
        'detachable': True,
        'attached': False
    }, _get_detached_key_id_list())

    return detached_ui_key_jnode_list


def get_attached_key_id_list():
    """
    :rtype: list of str
    """

    cmdline = ' '.join([_HASPRUS_FILEPATH, _HASPRUS_KEY_LIST])
    hasprus = subprocess.Popen(cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    hasprus_stdout, hasprus_stderr = hasprus.communicate()
    retcode = hasprus.returncode

    if retcode != 0:
        eventlog.log_e('fail: get attached keys list (%s, %s, %s)' % (hasprus_stdout, hasprus_stderr, retcode))
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    local_key_jnode_list = ujson.loads(hasprus_stdout)

    return local_key_jnode_list.keys()


def _get_detached_key_id_list():
    """
    :rtype: list of str
    """
    if _try_silent_mkdir_r(_DETACHED_KEYS_DIRPATH):
        return []

    detached_key_filepath_set = glob.glob(_DETACHED_KEYS_DIRPATH + '*.v2c')
    detached_key_id_list = map(
        lambda x: os.path.basename(x.split('_')[0]),
        detached_key_filepath_set)

    # make list distinct
    detached_key_id_list = list(set(detached_key_id_list))
    return detached_key_id_list


def get_demo_key_id():
    """
    Returns haspid of demokey if it is installed, returns None if demo-key not found
    :rtype: str | None
    """
    cmdline = ' '.join([_HASPRUS_FILEPATH, _HASPRUS_KEY_GETDEMO])
    hasprus = subprocess.Popen(cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    hasprus_stdout, hasprus_stderr = hasprus.communicate()
    retcode = hasprus.returncode

    if retcode != 0:
        eventlog.log_e('fail: get demo key id (%s, %s)' % (hasprus_stdout, hasprus_stderr))
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    demo_key_id = None
    demo_keys_list = ujson.loads(hasprus_stdout)

    if len(demo_keys_list) > 0:
        demo_key_id = demo_keys_list[0]

    return demo_key_id


def _handle_get_key_fingerprint():
    """
    handle ../key/fingerprint
    :rtype: dict
    """
    cmdline = ' '.join([_HASPRUS_FILEPATH, _HASPRUS_KEY_FINGERPRINT])
    hasprus = subprocess.Popen(cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    hasprus_stdout, hasprus_stderr = hasprus.communicate()
    retcode = hasprus.returncode

    if retcode == 0 and hasprus_stdout:
        return reply_done({
            'c2v': base64.b64encode(hasprus_stdout)
        })

    eventlog.log_e('fail: get fingerprint (%s, %s)' % (hasprus_stdout, hasprus_stderr))
    return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)


def _handle_post_key(v2c_content):
    """
    :type v2c_content: str
    """
    with open(_V2C_TEMP_FILEPATH, 'w') as license_file:
        license_file.write(v2c_content)

    cmdline = ' '.join([_HASPRUS_FILEPATH, _HASPRUS_KEY_UPDATE, _V2C_TEMP_FILEPATH])
    eventlog.log_d('try %s' % cmdline)
    hasprus = subprocess.Popen(cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    hasprus_stdout, hasprus_stderr = hasprus.communicate()
    retcode = hasprus.returncode
    if retcode != 0 or hasprus_stdout:
        raise RuntimeError('fail: activate key (%s, %s)' % (hasprus_stdout, hasprus_stderr))
    else:
        eventlog.log_d('done')


def apply_v2c_key(v2c_filepath):
    """
    :type v2c_filepath: str
    """
    cmdline = ' '.join([_HASPRUS_FILEPATH, _HASPRUS_KEY_UPDATE, v2c_filepath])
    eventlog.log_d('try %s' % cmdline)
    hasprus = subprocess.Popen(cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    hasprus_stdout, _ = hasprus.communicate()
    retcode = hasprus.returncode
    if retcode != 0 or hasprus_stdout:
        raise RuntimeError('fail: activate key %s' % hasprus_stdout)
    else:
        eventlog.log_d('done')


def _handle_get_key_all():
    attached_key_jnode = _get_attached_ui_key_jnode_list()
    detached_key_jnode = _get_detached_ui_key_jnode_list()
    return reply_done(attached_key_jnode + detached_key_jnode)


def _handle_post_key_id_attached(key_id):
    """
    :type key_id: str
    :rtype: dict
    """
    if key_id == 'embedded-demo':
        if get_demo_key_id():  # check demo-key used
            eventlog.log_d('demo already activated')
            return reply_done()

        # activate demo
        cmdline = ' '.join([_HASPRUS_FILEPATH, _HASPRUS_KEY_UPDATE, _DEMO_LICENSE_FILEPATH])
        eventlog.log_d('try %s' % cmdline)

        hasprus = subprocess.Popen(cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        hasprus_stdout, hasprus_stderr = hasprus.communicate()
        retcode = hasprus.returncode
        if retcode == 0 and not hasprus_stdout:
            eventlog.log_d('done')
            return reply_done()

        eventlog.log_e('fail: activate demo (%s, %s)' % (hasprus_stdout, hasprus_stderr))
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    # try attach other then 'embedded-demo' key
    if key_id not in _get_detached_key_id_list():
        eventlog.log_e('fail: not-found (requested key %s is absent)' % key_id)
        return reply_fail(REPLY_NOT_FOUND_404)

    _try_silent_mkdir_r(_DETACHED_KEYS_DIRPATH)

    key_to_attach_filepath_set = []

    key_to_attach_filepath = os.path.join(_DETACHED_KEYS_DIRPATH, key_id + _BASE_KEY_NAME_SUFFIX)
    if os.path.exists(key_to_attach_filepath):
        key_to_attach_filepath_set.append(key_to_attach_filepath)

    updated_key_to_attach_filepath_set = glob.glob(os.path.join(_DETACHED_KEYS_DIRPATH, key_id)
                                                   + _UPDATED_KEY_NAME_SUFFIX)
    for updated_key_to_attach in updated_key_to_attach_filepath_set:
        if os.path.exists(updated_key_to_attach):
            key_to_attach_filepath_set.append(updated_key_to_attach)

    if key_to_attach_filepath_set:
        _helpers.move_files(key_to_attach_filepath_set, _ATTACHED_KEYS_DIRPATH)
        _helpers.restart_hasp_service()
        return reply_done()


def _handle_delete_key_id_attached(key_id):
    """
    :type key_id: str
    :rtype: dict
    """
    _try_silent_mkdir_r(_DETACHED_KEYS_DIRPATH)

    key_to_detach_filepath_set = []

    key_to_detach_filepath = os.path.join(_ATTACHED_KEYS_DIRPATH, key_id + _BASE_KEY_NAME_SUFFIX)
    if os.path.exists(key_to_detach_filepath):
        key_to_detach_filepath_set.append(key_to_detach_filepath)

    updated_key_to_detach_filepath_set = glob.glob(os.path.join(_ATTACHED_KEYS_DIRPATH, key_id)
                                                   + _UPDATED_KEY_NAME_SUFFIX)
    for updated_key_to_detach in updated_key_to_detach_filepath_set:
        if os.path.exists(updated_key_to_detach):
            key_to_detach_filepath_set.append(updated_key_to_detach)

    if key_to_detach_filepath_set:
        _helpers.move_files(key_to_detach_filepath_set, _DETACHED_KEYS_DIRPATH)
        _helpers.restart_hasp_service()
        return reply_done()

    eventlog.log_e('fail: detach key')
    return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)


def _handle_delete_key_id(key_id):
    """
    :type key_id: str
    :rtype: dict
    """
    key_to_delete_filepath_set = []
    attached_key = False

    # attached "base" keys
    key_to_delete_filepath = os.path.join(_ATTACHED_KEYS_DIRPATH, key_id + _BASE_KEY_NAME_SUFFIX)
    if os.path.exists(key_to_delete_filepath):
        key_to_delete_filepath_set.append(key_to_delete_filepath)
        attached_key = True

    # attached "updated" keys
    updated_key_to_delete_filepath_set = glob.glob(os.path.join(_ATTACHED_KEYS_DIRPATH, key_id)
                                                   + _UPDATED_KEY_NAME_SUFFIX)

    for updated_key_to_delete in updated_key_to_delete_filepath_set:
        if os.path.exists(updated_key_to_delete):
            key_to_delete_filepath_set.append(updated_key_to_delete)
            attached_key = True

    # detached "base" key
    key_to_delete_filepath = os.path.join(_DETACHED_KEYS_DIRPATH, key_id + _BASE_KEY_NAME_SUFFIX)
    if os.path.exists(key_to_delete_filepath):
        key_to_delete_filepath_set.append(key_to_delete_filepath)

    # detached "updated" keys
    updated_key_to_delete_filepath_set = glob.glob(os.path.join(_DETACHED_KEYS_DIRPATH, key_id)
                                                   + _UPDATED_KEY_NAME_SUFFIX)
    for updated_key_to_delete in updated_key_to_delete_filepath_set:
        if os.path.exists(updated_key_to_delete):
            key_to_delete_filepath_set.append(updated_key_to_delete)

    # delete keys
    for filepath_to_delete in key_to_delete_filepath_set:
        if os.path.exists(filepath_to_delete):
            os.remove(filepath_to_delete)

        else:
            eventlog.log_e('fail: internal-server-error (file absent)')
            return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    if attached_key:
        _helpers.restart_hasp_service()
    return reply_done()


def _handle_get_key_id_fingerprint(key_id):
    """
    handle GET ../key/<id>/fingerprint
    :type key_id: str
    :rtype: dict
    """
    # check if key-id valid
    if not filter(lambda x: x['key_id'] == key_id, _vpw.lsvpwc(by_group=False)):
        eventlog.log_e('fail: key-id %s not found' % key_id)
        return reply_fail(REPLY_NOT_FOUND_404)

    cmdline = ' '.join([_HASPRUS_FILEPATH, _HASPRUS_KEY_INFO, key_id])
    hasprus = subprocess.Popen(cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    hasprus_stdout, hasprus_stderr = hasprus.communicate()
    retcode = hasprus.returncode

    if retcode == 0 and hasprus_stdout:
        return reply_done({
            'c2v': base64.b64encode(hasprus_stdout)
        })

    eventlog.log_e('fail: get fingerprint (%s, %s)' % (hasprus_stdout, hasprus_stderr))
    return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)


def _handle_get_key_id_group_all(key_id):
    # if license present but expired, the lsvpwc return licenses with
    # max_fps==0 and nthreads_max==0 and templates==''

    vpw_license_group_jnode_list = _vpw.lsvpwc(by_group=False)
    if key_id == _DEMO_KEY_ID:
        key_id = get_demo_key_id()

    filtered_key_set = filter(lambda x: x['key_id'] == key_id, vpw_license_group_jnode_list)
    if not filtered_key_set:
        eventlog.log_e('fail: key %s not found' % key_id)
        return reply_fail(REPLY_NOT_FOUND_404)

    d = defaultdict(lambda: defaultdict(set))

    for key in filtered_key_set:
        if d[key['fps_max']].get('templates', None) is None:
            d[key['fps_max']]['templates'] = key['templates'].split(' ')

        else:
            d[key['fps_max']]['templates'] += key['templates'].split(' ')

        if d[key['fps_max']].get('nthreads_max', None) is None:
            d[key['fps_max']]['nthreads_max'] = key['nthreads_max']

        else:
            d[key['fps_max']]['nthreads_max'] += key['nthreads_max']

    license_group_set = []

    for key, value in d.iteritems():
        license_group = {
            'id': str(key),
            'title': _FPS_TO_TITLE_MAP[key],
            'total-license-count': value['nthreads_max'],
            'fps-limit': key,
            'countries': _vpw.get_ui_countries_jnode(' '.join(value['templates']))
        }
        license_group_set.append(license_group)

    return reply_done(license_group_set)


def _handle_get_key_all_group_all():
    ui_license_group_jnode_list = map(lambda vpw_license_group_jnode: {
        'id': str(vpw_license_group_jnode['fps_max']),
        'title': _FPS_TO_TITLE_MAP[vpw_license_group_jnode['fps_max']],
        'total-license-count': vpw_license_group_jnode['nthreads_max'],
        'used-license-count': _helpers.get_used_license_count2(str(vpw_license_group_jnode['fps_max'])),
        'fps-limit': vpw_license_group_jnode['fps_max'],
        'countries': _vpw.get_ui_countries_jnode(vpw_license_group_jnode['templates'])
    }, _vpw.lsvpwc(by_group=True))

    # drop expired licenses from list
    expired_license_group_list = filter(lambda x: x['fps-limit'] == 0, ui_license_group_jnode_list)
    for expired in expired_license_group_list:
        ui_license_group_jnode_list.remove(expired)

    return reply_done(ui_license_group_jnode_list)


def _deserialize_key(query_args):
    """
    :type query_args: ui_types.UiQueryArgs
    :rtype: () -> dict
    """
    # POST [/api/v1/admin/actions/license/lpr/key]
    if query_args.http_method != 'POST':
        eventlog.log_e('fail: method-not-allowed (%s)' % query_args.http_method)
        return closure(reply_fail, REPLY_METHOD_NOT_ALLOWED_405)

    if query_args.query_string:
        eventlog.log_e('fail: bad-request (invalid query-string)')
        return closure(reply_fail, REPLY_BAD_REQUEST_400)

    if not query_args.raw_body:
        eventlog.log_e('fail: bad-request (invalid body)')
        return closure(reply_fail, REPLY_BAD_REQUEST_400)

    try:
        body_jnode = ujson.loads(query_args.raw_body)
        base64_v2c = body_jnode.get('v2c')
        if not base64_v2c:
            raise TypeError

        raw_v2c = base64.b64decode(base64_v2c)
        return closure(_handle_post_key, raw_v2c)

    except ValueError:
        eventlog.log_e('fail: bad-request (invalid body)')
        return closure(reply_fail, REPLY_BAD_REQUEST_400)

    except TypeError:
        eventlog.log_e('fail: bad-request (invalid body#/v2c)')
        return closure(reply_fail, REPLY_BAD_REQUEST_400)


def _deserialize_key_fingerprint(query_args):
    """
    :type query_args: ui_types.UiQueryArgs
    :rtype: () -> dict
    """
    path_segment_count = len(query_args.path_info)

    if path_segment_count == 1:
        # GET [/api/v1/admin/actions/license/lpr/key]/fingerprint
        if query_args.http_method != 'GET':
            eventlog.log_e('fail: method-not-allowed (%s)' % query_args.http_method)
            return closure(reply_fail, REPLY_METHOD_NOT_ALLOWED_405)

        if query_args.query_string:
            eventlog.log_e('fail: bad-request (invalid query-string)')
            return closure(reply_fail, REPLY_BAD_REQUEST_400)

        if query_args.raw_body:
            eventlog.log_e('fail: bad-request (invalid body)')
            return closure(reply_fail, REPLY_BAD_REQUEST_400)

        return closure(_handle_get_key_fingerprint)

    eventlog.log_e('fail: not-found')
    return closure(reply_fail, REPLY_NOT_FOUND_404)


def _deserialize_key_all(query_args):
    """
    :type query_args: ui_types.UiQueryArgs
    :rtype: () -> dict
    """
    path_segment_count = len(query_args.path_info)

    if path_segment_count == 1:
        # GET [/api/v1/admin/actions/license/lpr/key]/all
        if query_args.http_method != 'GET':
            eventlog.log_e('fail: method-not-allowed (%s)' % query_args.http_method)
            return closure(reply_fail, REPLY_METHOD_NOT_ALLOWED_405)

        if query_args.query_string:
            eventlog.log_e('fail: bad-request (invalid query-string)')
            return closure(reply_fail, REPLY_BAD_REQUEST_400)

        if query_args.raw_body:
            eventlog.log_e('fail: bad-request (wrong request body)')
            return closure(reply_fail, REPLY_BAD_REQUEST_400)

        return closure(_handle_get_key_all)

    elif path_segment_count == 3:
        # validate [/api/v1/admin/actions/license/lpr/key]/all/group/all
        _, raw_group, raw_group_id = query_args.path_info[0:3]
        if raw_group == 'group' and raw_group_id == 'all':
            return _deserialize_key_all_group_all(query_args)

    eventlog.log_e('fail: not-found')
    return closure(reply_fail, REPLY_NOT_FOUND_404)


def _deserialize_key_all_group_all(query_args):
    """
    :type query_args: ui_types.UiQueryArgs
    :rtype: () -> dict
    """
    # GET [/api/v1/admin/actions/license/lpr/key]/all/group/all
    if query_args.http_method != 'GET':
        eventlog.log_e('fail: method-not-allowed (%s)' % query_args.http_method)
        return closure(reply_fail, REPLY_METHOD_NOT_ALLOWED_405)

    if query_args.query_string:
        eventlog.log_e('fail: bad-request (invalid query-string)')
        return closure(reply_fail, REPLY_BAD_REQUEST_400)

    if query_args.raw_body:
        eventlog.log_e('fail: bad-request (wrong request body)')
        return closure(reply_fail, REPLY_BAD_REQUEST_400)

    return closure(_handle_get_key_all_group_all)


def _deserialize_key_id(query_args):
    """
    :type query_args: ui_types.UiQueryArgs
    :rtype: () -> dict
    """
    path_segment_count = len(query_args.path_info)

    if path_segment_count == 1:
        # DELETE [/api/v1/admin/actions/license/lpr/key]/<key-id>
        raw_key_id, = query_args.path_info

        if query_args.http_method != 'DELETE':
            eventlog.log_e('fail: method-not-allowed (%s)' % query_args.http_method)
            return closure(reply_fail, REPLY_METHOD_NOT_ALLOWED_405)

        if query_args.query_string:
            eventlog.log_e('fail: bad-request (invalid query-string)')
            return closure(reply_fail, REPLY_BAD_REQUEST_400)

        if query_args.raw_body:
            eventlog.log_e('fail: bad-request (wrong request body)')
            return closure(reply_fail, REPLY_BAD_REQUEST_400)

        return closure(_handle_delete_key_id, raw_key_id)

    elif path_segment_count == 2:
        raw_key_id, raw_action = query_args.path_info

        if raw_action == 'fingerprint':
            # [/api/v1/admin/actions/license/lpr/key]/<key-id>/fingerprint
            return _deserialize_key_id_fingerprint(query_args)

        if raw_action == 'attached':
            # [/api/v1/admin/actions/license/lpr/key]/<key-id>/attached
            return _deserialize_key_id_attached(query_args)

    elif path_segment_count == 3:
        # [/api/v1/admin/actions/license/lpr/key]/<key-id>/group/all
        raw_key_id, raw_group, raw_group_id = query_args.path_info
        if raw_group == 'group' and raw_group_id == 'all':
            return _deserialize_key_id_group_all(query_args)

    eventlog.log_e('fail: not-found')
    return closure(reply_fail, REPLY_NOT_FOUND_404)


def _deserialize_key_id_fingerprint(query_args):
    """
    :type query_args: ui_types.UiQueryArgs
    :rtype: () -> dict
    """
    # GET [/api/v1/admin/actions/license/lpr/key]/<key-id>/fingerprint
    key_id = query_args.path_info[0]

    if query_args.http_method != 'GET':
        eventlog.log_e('fail: method-not-allowed (%s)' % query_args.http_method)
        return closure(reply_fail, REPLY_METHOD_NOT_ALLOWED_405)

    if query_args.query_string:
        eventlog.log_e('fail: bad-request (invalid query-string)')
        return closure(reply_fail, REPLY_BAD_REQUEST_400)

    if query_args.raw_body:
        eventlog.log_e('fail: bad-request (invalid body)')
        return closure(reply_fail, REPLY_BAD_REQUEST_400)

    return closure(_handle_get_key_id_fingerprint, key_id)


def _deserialize_key_id_attached(query_args):
    """
    :type query_args: ui_types.UiQueryArgs
    :rtype: () -> dict
    """
    # POST   [/api/v1/admin/actions/license/lpr/key]/<key-id>/attached
    # DELETE [/api/v1/admin/actions/license/lpr/key]/<key-id>/attached
    if query_args.http_method == 'POST':
        return _deserialize_post_key_id_attached(query_args)

    elif query_args.http_method == 'DELETE':
        return _deserialize_delete_key_id_attached(query_args)

    eventlog.log_e('fail: method-not-allowed (%s)' % query_args.http_method)
    return closure(reply_fail, REPLY_METHOD_NOT_ALLOWED_405)


def _deserialize_post_key_id_attached(query_args):
    """
    :type query_args: ui_types.UiQueryArgs
    :rtype: () -> dict
    """
    key_id = query_args.path_info[0]

    if query_args.query_string:
        eventlog.log_e('fail: bad-request (invalid query-string)')
        return closure(reply_fail, REPLY_BAD_REQUEST_400)

    if query_args.raw_body:
        body_jnode = ujson.loads(query_args.raw_body)
        if body_jnode:  # TODO: fix {} in POST-request from frontend
            eventlog.log_e('fail: bad-request (invalid body)')
            return closure(reply_fail, REPLY_BAD_REQUEST_400)

    return closure(_handle_post_key_id_attached, key_id)


def _deserialize_delete_key_id_attached(query_args):
    """
    :type query_args: ui_types.UiQueryArgs
    :rtype: () -> dict
    """
    key_id = query_args.path_info[0]
    if key_id not in get_attached_key_id_list():
        eventlog.log_e('fail: not-found (key %s is absent)' % key_id)
        return closure(reply_fail, REPLY_NOT_FOUND_404)

    if query_args.query_string:
        eventlog.log_e('fail: bad-request (invalid query-string)')
        return closure(reply_fail, REPLY_BAD_REQUEST_400)

    if query_args.raw_body:
        eventlog.log_e('fail: bad-request (invalid body)')
        return closure(reply_fail, REPLY_BAD_REQUEST_400)

    return closure(_handle_delete_key_id_attached, key_id)


def _deserialize_key_id_group_all(query_args):
    """
    :type query_args: ui_types.UiQueryArgs
    :rtype: () -> dict
    """
    # GET [/api/v1/admin/actions/license/lpr/key]/<key-id>/group/all
    key_id = query_args.path_info[0]

    if query_args.http_method != 'GET':
        eventlog.log_e('fail: method-not-allowed (%s)' % query_args.http_method)
        return closure(reply_fail, REPLY_METHOD_NOT_ALLOWED_405)

    if query_args.query_string:
        eventlog.log_e('fail: bad-request (invalid query-string)')
        return closure(reply_fail, REPLY_BAD_REQUEST_400)

    if query_args.raw_body:
        eventlog.log_e('fail: bad-request (wrong request body)')
        return closure(reply_fail, REPLY_BAD_REQUEST_400)

    return closure(_handle_get_key_id_group_all, key_id)


def _main():
    eventlog.open_log('cgi-license', 'debug')
    exit_code = 1
    # noinspection PyBroadException
    try:
        query_args = ui_types.UiQueryArgs.deserialize()

        if len(query_args.path_info) == 0:
            # [/api/v1/admin/actions/license/lpr/key]
            handler = _deserialize_key(query_args)

        elif query_args.path_info[0] == 'fingerprint':
            # [/api/v1/admin/actions/license/lpr/key]/fingerprint
            handler = _deserialize_key_fingerprint(query_args)

        elif query_args.path_info[0] == 'all':
            # [/api/v1/admin/actions/license/lpr/key]/all
            handler = _deserialize_key_all(query_args)

        else:
            # [/api/v1/admin/actions/license/lpr/key]/<key-id>
            handler = _deserialize_key_id(query_args)

        reply_jnode = handler()
        exit_code = 0

    except:
        reply_jnode = reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)
        eventlog.log_x()

    finally:
        eventlog.close_log()

    print ujson.dumps(reply_jnode)
    sys.exit(exit_code)


if __name__ == '__main__':
    _main()
