import ujson
import os

from edge.admin.ui_types import BaseCreator
from edge.path_helpers import *
from os.path import join as join_path

import edge.pipeline_helpers as helpers


def rg_try_get_creator_ctor(ui_entry_point, used_ui_modules, used_edge_modules, avaliable_ui_modules):
    """
    :type ui_entry_point: UiIdentity
    :type used_ui_modules: set of UiIdentity
    :type used_edge_modules: set of EdgeIdentity
    :type avaliable_ui_modules: set of UiIdentity
    :rtype: None | (str, UiIdentity, dict of UiIdentity: dict) -> BaseCreator
    """
    if ui_entry_point.ptype != PTYPE_LPR_RTSP:
        return None

    ui_modules_set = set()
    if ui_entry_point.type in [UI_TYPE_VIDEOSOURCE_RTSP, UI_TYPE_RED_LIGHT, UI_TYPE_PIPELINE]:
        for ui_module in avaliable_ui_modules:
            if ui_module.type in [UI_TYPE_VIDEOSOURCE_RTSP, UI_TYPE_RED_LIGHT]:
                ui_modules_set.add(ui_module)

        if ui_entry_point.type != UI_TYPE_PIPELINE:
            used_edge_modules.add(EdgeIdentity(EDGE_TYPE_RTSP_GRABBER, '0'))

        used_ui_modules.update(ui_modules_set)
        creator_ctor = RgCreator.create
    else:
        return None

    return lambda arg0, arg1, arg2: creator_ctor(arg0, arg1, arg2, avaliable_ui_modules)


class RgCreator(BaseCreator):
    def __init__(self, id_main_setupstore_halt):
        """
        :type id_main_setupstore_halt: str, dict, str, str
        """
        BaseCreator.__init__(self)

        self._id_main_setupstore_halt = id_main_setupstore_halt

    @staticmethod
    def create(ui_pid, ui_entry_point, ui_identity_to_jnode_map, avaliable_ui_modules):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :type avaliable_ui_modules: set of UiIdentity
        :rtype: RgCreator
        """

        if ui_entry_point.type in [UI_TYPE_VIDEOSOURCE_RTSP, UI_TYPE_RED_LIGHT, UI_TYPE_PIPELINE]:
            id_main_setupstore_halt = RgCreator._serialize(ui_pid, ui_identity_to_jnode_map, avaliable_ui_modules)
            creator = RgCreator(id_main_setupstore_halt)

        else:
            raise NotImplemented('entry_point not supported')

        return creator

    @staticmethod
    def _serialize(ui_pid, ui_identity_to_jnode_map, avaliable_ui_modules):
        """
        :type ui_pid: str
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :type avaliable_ui_modules: set of UiIdentity
        :rtype: str, dict, str
        """
        ui_videosource_jnode = ui_identity_to_jnode_map[Pipeline.Rtsp.Ui.VIDEOSOURCE_0]

        use_udp = False
        transport = ui_videosource_jnode['rw']['conf']['video'].get('transport', ["RTP/AVP/UDP", "RTP/AVP/TCP"])
        if "RTP/AVP/UDP" in transport:
            use_udp = True

        skip_rate = ui_videosource_jnode['rw']['conf']['video'].get('skip-rate', 1) - 1

        spool_dir = get_spool_dir(ui_pid)
        edge_group = helpers.get_edge_group(ui_pid)
        edge_gid = Pipeline.Rtsp.Edge.RTSPGRABBER_0.gid

        ui_format_to_edge_format = {
            'video/H264': 'h264',
            'video/JPEG': 'jpeg'
        }
        use_gpu = 'hw/auto' == ui_videosource_jnode['rw']['conf']['video'].get('decoder', 'hw/auto')
        # create main.conf
        main_conf_jnode = {
            "InputStreamType": ui_format_to_edge_format.get(ui_videosource_jnode['rw']['conf']['video']['format']),
            "UseUDP": use_udp,
            "SkipRate": skip_rate,
            "TargetHeight": 0,
            "Url": ui_videosource_jnode['rw']['conf']['url'],
            "TargetWidth": 0,
            "UseGpu": use_gpu,
            "Store": {
                "Directory": join_path(spool_dir, 'rg%s-rg%s-store' % (edge_gid, edge_gid)),
                "FileCountLow": 40,
                "FileCountHigh": 45
            },
            "FileApi": {
                "SceneEventMono8t": [],
                "SceneEventRaw": [],
                "HaltEvent": {
                    "SignalAbsentTout": 5000,
                    "DeviceAbsentTout": 600000
                },
            },
            "Monitoring": {
                "StatTout": 5,
                "StatPath": join_path(spool_dir, 'zabbix', 'vs%s' % edge_gid)
            }
        }
        rgbs = ui_videosource_jnode['ro'].get("rgbs", False)
        mcgst_scene = ui_videosource_jnode['ro'].get("mcgst-scene", True)

        if rgbs:
            edge_identity = Pipeline.Rtsp.Edge.RTSPGRABBER_0
            edge_id = helpers.get_edge_id(edge_group, edge_identity)
            main_conf_jnode['StoreLbsv'] = [edge_id, '0']

        color_mode_grayscale = ui_videosource_jnode['rw']['conf']['video'].get('grayscale', False)
        if color_mode_grayscale:
            main_conf_jnode['FileApi']['SceneEventMono8t'].append(
                join_path(spool_dir, 'rg%s-lr%s-scene' % (edge_gid, edge_gid)))

            main_conf_jnode['FileApi']['SceneEventMono8t'].append(
                join_path(spool_dir, 'rg%s-tr%s-scene' % (edge_gid, edge_gid)))

            if mcgst_scene and not rgbs:
                main_conf_jnode['FileApi']['SceneEventMono8t'].append(
                    join_path(spool_dir, 'mg%s-mcgst%s-scene' % (edge_gid, edge_gid)))

        else:
            main_conf_jnode['FileApi']['SceneEventRaw'].append(
                join_path(spool_dir, 'rg%s-lr%s-scene' % (edge_gid, edge_gid)))

            main_conf_jnode['FileApi']['SceneEventRaw'].append(
                join_path(spool_dir, 'rg%s-tr%s-scene' % (edge_gid, edge_gid)))

            if mcgst_scene and not rgbs:
                main_conf_jnode['FileApi']['SceneEventRaw'].append(
                    join_path(spool_dir, 'mg%s-mcgst%s-scene' % (edge_gid, edge_gid)))

        # add mg to rl path
        ui_redlight_rules_id_to_jnode_list = []
        for ui_module in avaliable_ui_modules:
            if ui_module.type == UI_TYPE_RED_LIGHT:
                ui_redlight_rules_id_to_jnode_list.append((ui_module.id, ui_identity_to_jnode_map[ui_module]))

        if len(ui_redlight_rules_id_to_jnode_list) == 1:
            ui_id, red_light_jnode = ui_redlight_rules_id_to_jnode_list[0]
            vendor = red_light_jnode['ro'].get('vendor', 'zonechecker')
            if vendor == 'zonechecker':
                main_conf_jnode['FileApi']['SceneEventMono8t'].append(
                    join_path(spool_dir, 'mg%s-zc%s-scene' % (edge_gid, edge_gid))
                )

        # create setupstore.conf
        edge_id = helpers.get_edge_id(edge_group, Pipeline.Rtsp.Edge.RTSPGRABBER_0)
        setupstore_conf_text = get_setupstore_conf_text(edge_id)

        # create haltaction.conf
        def_halt_list = [
            "0:python /opt/edge/bin/psuctl run led-camera color G blink 0HZ",
            "1:python /opt/edge/bin/psuctl run led-camera color R blink 1HZ",
            "2:python /opt/edge/bin/psuctl run led-camera color R blink 0HZ"
        ]
        halt_list = ui_videosource_jnode['ro'].get('haltaction', def_halt_list)
        if halt_list:
            haltaction_conf = '\n'.join(halt_list)

        else:
            haltaction_conf = ''

        return edge_id, main_conf_jnode, setupstore_conf_text, haltaction_conf

    def serialize(self):
        edge_id, main_conf_jnode, setupstore_conf_text, haltaction_conf_text = \
            self._id_main_setupstore_halt

        edge_jnode_dir = get_edge_jnode_dirpath(EDGE_TYPE_RTSP_GRABBER, edge_id)
        if not os.path.exists(edge_jnode_dir):
            helpers.silent_mkdir_r(edge_jnode_dir)

        main_conf_abspath = join_path(edge_jnode_dir, 'main.conf')
        with open(main_conf_abspath, 'w') as main_conf:
            ujson.dump(main_conf_jnode, main_conf, indent=2)

        setupstore_conf_path = join_path(edge_jnode_dir, 'setupstore.conf')

        with open(setupstore_conf_path, 'w') as text_file:
            text_file.write(setupstore_conf_text)

        halt_action_path = join_path(edge_jnode_dir, 'haltaction.conf')
        with open(halt_action_path, 'w') as text_file:
            text_file.write(haltaction_conf_text)


class RgDeleter:
    def __init__(self, edge_rg_dirpath):
        """
        :type edge_rg_dirpath: str
        """
        self._edge_rg_dirpath = edge_rg_dirpath

    @staticmethod
    def create(ui_pid, ui_entry_point):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :rtype: RgDeleter
        """
        edge_group = helpers.get_edge_group(ui_pid)

        if ui_entry_point == Pipeline.Rtsp.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Rtsp.Edge.RTSPGRABBER_0)
            edge_type = Pipeline.Rtsp.Edge.RTSPGRABBER_0.type
            deleter = RgDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        elif ui_entry_point == Pipeline.Mv.Ui.PIPELINE:
            return None

        elif ui_entry_point == Pipeline.Speed.Ui.PIPELINE:
            return None

        else:
            raise NotImplemented

    def run(self):
        helpers.silent_rm_r(self._edge_rg_dirpath)
