import jsonschema
import ujson

from edge import eventlog, path_helpers as ph, pipeline_helpers as oh
from edge.admin import reply_helpers as rh, ui_types as ui
from edge.com import fs


import edge_creator as creator
from ui_pipeline_set import get_pipeline_type
from copy import deepcopy


def _check_rtsp_link_changed(old_ui_jnode, new_ui_jnode):
    return old_ui_jnode['conf']['url'] != new_ui_jnode['conf']['url']


def _check_rtsp_changed(old_ui_jnode, new_ui_jnode):
    if old_ui_jnode['conf']['url'] != new_ui_jnode['conf']['url']:
        return True

    if old_ui_jnode['conf']['video']['transport'] != new_ui_jnode['conf']['video']['transport']:
        return True

    if old_ui_jnode['conf']['video'].get('decoder') != new_ui_jnode['conf']['video'].get('decoder'):
        return True

    if old_ui_jnode['conf']['video'].get('format') != new_ui_jnode['conf']['video'].get('format'):
        return True

    if old_ui_jnode['conf']['video'].get('color-mode') != new_ui_jnode['conf']['video'].get('color-mode'):
        return True

    if old_ui_jnode['conf']['video'].get('skip-rate') != new_ui_jnode['conf']['video'].get('skip-rate'):
        return True

    return False


def _handle_get(ui_pid, entry_point):
    """
    :type ui_pid: str
    :type entry_point: ph.UiIdentity
    """
    ui_jnode_path = ph.get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)
    ui_jnode = fs.load_json2(ui_jnode_path)
    ui_jnode_rw = ui_jnode.get('rw', dict())

    stats_jnode = oh.det_videosource_stats(ui_pid, entry_point.id)
    if stats_jnode:
        ui_jnode_rw['stats'] = stats_jnode

    return rh.reply_done(ui_jnode_rw, envelope=False)


def _put_rtsp(ui_pid, entry_point, new_ui_jnode):
    """
    :type ui_pid: str
    :type entry_point: ph.UiIdentity
    :type new_ui_jnode: dict
    :rtype: dict
    """
    restart_blobstore = []

    def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
        """
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :return: dict
        """
        if not _check_rtsp_changed(ui_identity_to_jnode_map[entry_point]['rw'], new_ui_jnode):
            return False

        # if link changed:
        # restart blobstore with video archive to clear h264 buffer:
        if _check_rtsp_link_changed(ui_identity_to_jnode_map[entry_point]['rw'], new_ui_jnode):
            restart_blobstore.append(True)

        ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode

        return True

    try:
        creator.generate(ui_pid, entry_point, _edit_ui_jnode_handler)

        if restart_blobstore:
            edge_id = oh.get_edge_id(ui_pid, ph.EdgeIdentity(ph.EDGE_TYPE_BLOBSTORE, '0'))
            oh.sighup_edge_module(ph.EDGE_TYPE_BLOBSTORE, edge_id)

        return rh.reply_done(new_ui_jnode, envelope=False)

    except ui.LockError:
        return rh.reply_fail(rh.REPLY_CONFLICT_409, envelope=False)


class UiRtspVideosource:
    def __init__(self, handler):
        """
        :type handler: object -> dict
        """
        self._handler = handler

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiRtspVideosource
        """
        #
        http_method = query_args.http_method
        url_query_string = query_args.query_string
        body_jnode = ujson.loads(query_args.raw_body) if query_args.raw_body else None

        ui_pid = query_args.get_ui_pid()
        ui_type = query_args.get_ui_type()
        ui_id = query_args.get_ui_id()
        if None in [ui_pid, ui_type, ui_id]:
            return UiRtspVideosource(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404, envelope=False))

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype != ph.PTYPE_LPR_RTSP:
            return UiRtspVideosource(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400, envelope=False))

        if ui_type != ph.UI_TYPE_VIDEOSOURCE_RTSP:
            return UiRtspVideosource(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400, envelope=False))

        entry_point = ph.UiIdentity(ui_ptype, ui_type, ui_id)

        #
        if url_query_string:
            eventlog.log_d('fail: wrong query-string')
            return UiRtspVideosource(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400, envelope=False))

        #
        if http_method == 'GET':
            if not body_jnode:
                return UiRtspVideosource(lambda: _handle_get(ui_pid, entry_point))

        elif http_method == 'PUT':
            schema_abs_path = ph.get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            request_schema_jnode = fs.load_json2(schema_abs_path)
            if body_jnode:
                try:
                    jsonschema.validate(body_jnode, request_schema_jnode)
                    return UiRtspVideosource(lambda: _put_rtsp(ui_pid, entry_point, body_jnode))

                except (ValueError, jsonschema.ValidationError) as x:
                    eventlog.log_d('fail: %s' % x.message)
                    return UiRtspVideosource(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400, envelope=False))

        else:
            eventlog.log_d('fail: method not allowed')
            return UiRtspVideosource(lambda: rh.reply_fail(rh.REPLY_METHOD_NOT_ALLOWED_405, envelope=False))

        eventlog.log_d('fail: deserialize')
        return UiRtspVideosource(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400, envelope=False))

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler()
