__all__ = [
    'get_openvpn_info',
    'add_profile',
    'delete_profile',
    'make_profile_active',
    'NotFoundError',
    'BadRequestError'
]


from edge.com import fs
from edge import eventlog
from edge.com import sh


class NotFoundError(RuntimeError):
    pass


class BadRequestError(RuntimeError):
    pass


JNODE_PATH = '/etc/edge/setupstore/admin/actions/openvpn.json'
ETC_DEFAULT_OPENVPN_FILEPATH = '/etc/default/openvpn'


def _get_new_ref(jnode):
    """
    :type jnode: dict
    :rtype: str | None
    """
    used_ref_set = set(map(lambda profile_jnode: int(profile_jnode['ref']), jnode['rw']['conf']['profiles']))

    ref = None
    for ref_candidate in xrange(0, 32):
        if ref_candidate not in used_ref_set:
            ref = str(ref_candidate)
            break

    if ref is None:
        eventlog.log_e("fail: get new-ref")
    return ref


def _openvpn_reload_restart():
    sh.run_silent('service openvpn stop')
    sh.run_silent('systemctl daemon-reload')
    sh.run_silent('service openvpn start')


def _mk_profile_path(ref):
    """
    :type ref: str
    :rtype: str
    """
    return '/etc/openvpn/{}.conf'.format(ref)


def _mk_user_path(ref):
    """
    :type ref: str
    :rtype: str
    """
    return '/etc/openvpn/{}.user'.format(ref)


def _get_active_profiles(jnode):
    """
    :type jnode: dict
    :rtype: set[str]
    """
    active_ref_set = set()
    for profile_jnode in jnode['rw']['conf']['profiles']:
        if profile_jnode['stats']['active']:
            active_ref_set.add(profile_jnode['ref'])

    return active_ref_set


def _set_autostart_profile(active_profiles):
    """
    :type active_profiles: set[str]
    :rtype: int
    """
    # Edit profile autostart.
    with open(ETC_DEFAULT_OPENVPN_FILEPATH, 'r') as old_default_vpn_file:
        old_defaults_vpn_lines = old_default_vpn_file.readlines()

    new_defaults_vpn_lines = []
    for old_line in old_defaults_vpn_lines:
        if not old_line.startswith('AUTOSTART='):
            new_defaults_vpn_lines.append(old_line)

    if len(active_profiles) == 0:
        autostart_line = 'AUTOSTART="none"\n'

    else:
        autostart_line = 'AUTOSTART="%s"\n' % ' '.join(active_profiles)

    new_defaults_vpn_lines.append(autostart_line)

    with open(ETC_DEFAULT_OPENVPN_FILEPATH, 'w') as new_default_vpn_file:
        new_default_vpn_file.write(''.join(new_defaults_vpn_lines))


def _add_auth_to_profile(conf_path, user_path, prof_user=None, prof_pass=None):
    """
    :type conf_path: str
    :type prof_user: str | None
    :type prof_pass: str | None
    :raises: BadRequestError
    """
    with open(conf_path, 'r') as old_default_vpn_file:
        old_defaults_vpn_lines = old_default_vpn_file.readlines()

    new_defaults_vpn_lines = []
    auth_user_path = False
    for old_line in old_defaults_vpn_lines:
        if old_line == 'auth-user-pass\n':
            auth_user_path = True
            if None in [prof_user, prof_pass]:
                eventlog.log_e('fail: user/pass not defined')
                raise BadRequestError

            auth_conf = 'auth-user-pass {}\n'.format(user_path)
            new_defaults_vpn_lines.append(auth_conf)

        else:
            new_defaults_vpn_lines.append(old_line)

    if auth_user_path:
        with open(conf_path, 'w') as new_default_vpn_file:
            new_default_vpn_file.write(''.join(new_defaults_vpn_lines))

        with open(user_path, 'w') as user_vpn_file:
            user_vpn_file.write('{}\n{}\n'.format(prof_user, prof_pass))


def get_openvpn_info():
    """
    :rtype: dict | None
    """
    jnode = fs.load_json(JNODE_PATH)

    if jnode is None:
        raise Exception
    return jnode.get('rw')


def add_profile(name, conf_path, prof_user=None, prof_pass=None):
    """
    :type name: str
    :type conf_path: str
    :type prof_user: str | None
    :type prof_pass: str | None
    :rtype: dict
    :raises: BadRequestError
    """
    # TODO: validate name is unique
    openvpn_jnode = fs.load_json(JNODE_PATH)

    if len(filter(lambda _jnode: _jnode['name'] == name, openvpn_jnode['rw']['conf']['profiles'])):
        eventlog.log_e('fail: name {} already exist'.format(name))
        raise BadRequestError

    new_profile_ref = _get_new_ref(openvpn_jnode)

    user_path = _mk_user_path(new_profile_ref)
    # check if auth-user-path and user + pass not None
    _add_auth_to_profile(conf_path, user_path, prof_user, prof_pass)

    new_profile_path = _mk_profile_path(new_profile_ref)
    fs.silent_cp(conf_path, new_profile_path)

    new_profile_jnode = {
        "ref": new_profile_ref,
        "name": name,
        "internal": False,
        "stats": {
            "active": False
        }
    }

    openvpn_jnode['rw']['conf']['profiles'].append(new_profile_jnode)
    fs.write_json_file(openvpn_jnode, JNODE_PATH, True)
    return openvpn_jnode.get('rw')


def delete_profile(ref):
    """
    :type ref: str
    :rtype: dict
    :raises: BadRequestError, NotFoundError
    """
    # TODO: check if profile exist
    # TODO: check if profile internal
    profile_path = _mk_profile_path(ref)
    user_path = _mk_user_path(ref)

    fs.silent_rm_r(profile_path)
    fs.silent_rm_r(user_path)

    openvpn_jnode = fs.load_json(JNODE_PATH)

    new_profiles = filter(lambda profile_jnode: profile_jnode['ref'] != ref, openvpn_jnode['rw']['conf']['profiles'])

    if len(new_profiles) == len(openvpn_jnode['rw']['conf']['profiles']):
        eventlog.log_e('fail: profile {} not exist'.format(ref))
        raise NotFoundError

    profile_to_delete = filter(lambda _jnode: _jnode['ref'] == ref, openvpn_jnode['rw']['conf']['profiles'])[0]
    if profile_to_delete['internal']:
        eventlog.log_e('fail: profile {} is internal'.format(ref))
        raise BadRequestError

    openvpn_jnode['rw']['conf']['profiles'] = new_profiles

    active_profiles = _get_active_profiles(openvpn_jnode)
    _set_autostart_profile(active_profiles)
    _openvpn_reload_restart()

    fs.write_json_file(openvpn_jnode, JNODE_PATH, True)
    return openvpn_jnode.get('rw')


def make_profile_active(ref, active):
    """
    :type ref: str
    :type active: bool
    :rtype: dict
    :raises: NotFoundError
    """
    openvpn_jnode = fs.load_json(JNODE_PATH)
    profiles_jnode = filter(lambda _jnode: _jnode['ref'] == ref, openvpn_jnode['rw']['conf']['profiles'])

    if len(profiles_jnode) == 0:
        eventlog.log_e('fail: profile {} nor exist'.format(ref))
        raise NotFoundError
    elif len(profiles_jnode) > 1:
        raise Exception

    profile_jnode = profiles_jnode[0]
    profile_jnode['stats']['active'] = active

    active_profiles = _get_active_profiles(openvpn_jnode)
    _set_autostart_profile(active_profiles)
    _openvpn_reload_restart()

    fs.write_json_file(openvpn_jnode, JNODE_PATH, True)
    return openvpn_jnode.get('rw')


